﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using IdeaBlade.EntityModel;
using DomainModel;
using System.IO;
using IdeaBlade.Core;

using Crypto = IdeaBlade.Core.CryptoFns;

namespace Console01 {
  class Program {
    static void Main(string[] args) {
      new Program().DoIt();
    }

    private void DoIt() {
      LoadDataIntoCache();
      
      string cacheFilePath = System.IO.Directory.GetCurrentDirectory() + "\\devForceCache.bin";
      byte[] key = Encrypto.TestHashKey;

      WriteCache(cacheFilePath, key);
      ClearCache();
      ReadCacheAndDisplayContents(cacheFilePath, key);

      PromptToContinue();
    }

    private void ReadCacheAndDisplayContents(string cacheFilePath, byte[] key) {
      Console.WriteLine("Restoring cache state from {0}", cacheFilePath);
      ImportCacheStateFile(cacheFilePath, key);
      Console.WriteLine("\nCache contents:\n");
      DisplayCacheContents();
    }

    private void ClearCache() {
      Console.WriteLine("\nClearing cache...\n");
      _mgr.Clear();
    }

    private void WriteCache(string cacheFilePath, byte[] key) {
      Console.WriteLine("Saving encrypted cache state to {0}", cacheFilePath);
      SaveCacheToCacheStateFile(cacheFilePath, key);
    }

    private void SaveCacheToCacheStateFile(string cacheFilePath, byte[] key) {
      byte[] encryptedData;
      using (var cacheStateStream = new MemoryStream()) {
        _mgr.CacheStateManager.SaveCacheState(cacheStateStream, true);
        encryptedData = Encrypto.Encrypt(key, cacheStateStream);
        WriteByteArrayToFile(encryptedData, cacheFilePath);
      }
    }

    private void ImportCacheStateFile(string cacheFilePath, byte[] key) {
      byte[] encryptedData;
      byte[] decryptedData;

      encryptedData = ReadByteArrayFromFile(cacheFilePath);
      decryptedData = Encrypto.Decrypt(key, encryptedData);

      bool restoreSaveOptions = true;
      bool restoreQueryStrategy = false;
      var cacheRestoreStrategy = new RestoreStrategy(
        restoreSaveOptions, restoreQueryStrategy, MergeStrategy.OverwriteChanges);

      try {
        _mgr.CacheStateManager.RestoreCacheState(
          new MemoryStream(decryptedData), cacheRestoreStrategy, true);
      }
      catch (Exception e) {
        Console.WriteLine(e.Message);
      }
    }


    private void DoIt01() {
      LoadDataIntoCache();
      Console.WriteLine("Entities present after initial retrieval: {0}", GetCountOfEntitiesInCache());
      StoreCacheStateLocally();
      _mgr.Clear();
      Console.WriteLine("Entities present after Clear and encrypted local save: {0}", GetCountOfEntitiesInCache());
      RestoreLocallyStoredCacheState();
      Console.WriteLine("Entities present after restore of encrypted local file: {0}", GetCountOfEntitiesInCache());
      PromptToContinue();
    }

    private string _encryptionString = "EncryptMe";
    private byte[] _encryptionKey;
    private string _targetPath = "c:\\DevForceCache.bin";

    private void LoadDataIntoCache() {
      List<Customer> customers = new List<Customer>(
        _mgr.Customers.Include("Orders").Include("Orders.OrderDetails").ToList());
    }

    private void StoreCacheStateLocally() {
      _encryptionKey = CryptoFns.AesGetHashKey(_encryptionString);

      bool fipsCompliant = true;
      string binaryCacheDataAsString = null;
      string encryptedBinaryCacheDataAsString = null;
      MemoryStream aMemoryStream = new MemoryStream();
      StreamWriter aStreamWriter = new StreamWriter(_targetPath, false);

      try {
        _mgr.CacheStateManager.SaveCacheState(aMemoryStream, true);
        binaryCacheDataAsString = Convert.ToBase64String(aMemoryStream.GetBuffer());
        encryptedBinaryCacheDataAsString = IdeaBlade.Core.CryptoFns.AesEncrypt(binaryCacheDataAsString, _encryptionKey, fipsCompliant);

        // BEG Debugging variables
        string cacheString01 = binaryCacheDataAsString.Substring(0, 100);
        string encryptedCacheString01 = encryptedBinaryCacheDataAsString.Substring(0, 100);
        string decryptedCacheString01 = Crypto.AesDecrypt(encryptedBinaryCacheDataAsString, _encryptionKey, true).Substring(0, 100);
        // END Debugging variables

        aStreamWriter.Write(encryptedBinaryCacheDataAsString);
      }
      catch (Exception ex) {
        Console.WriteLine(ex.Message);
      }
      finally {
        aStreamWriter.Flush();
        aStreamWriter.Close();
        string msg = "Local cache contents saved in encrypted form to local storage at: ";
        Console.WriteLine(msg + _targetPath);
      }
    }

    private void RestoreLocallyStoredCacheState() {
      _encryptionKey = CryptoFns.AesGetHashKey(_encryptionString);
      MemoryStream aMemoryStream = null;
      string encryptedBinaryCacheDataAsString = null;
      byte[] binaryCacheData = null;

      try {
        //Read the file
        StreamReader aStreamReader = new StreamReader(_targetPath);
        try {
          //Convert the encrypted data to a string
          encryptedBinaryCacheDataAsString = aStreamReader.ReadToEnd();

          // BEG debugging variables
          string encryptedCacheString02 = encryptedBinaryCacheDataAsString.Substring(0, 100);
          // END debugging variables

          //Decrypt the encrypted string
          string decryptedBinaryCacheDataAsString =
            IdeaBlade.Core.CryptoFns.AesDecrypt(encryptedBinaryCacheDataAsString, _encryptionKey, true);

          // BEG debugging variables
          string decryptedCacheString02 = decryptedBinaryCacheDataAsString.Substring(0, 100);
          // END debugging variables

          binaryCacheData = Convert.FromBase64String(decryptedBinaryCacheDataAsString);

          //Read the decrypted string into a MemoryStream
          aMemoryStream = new MemoryStream(binaryCacheData);
          aMemoryStream.Position = 0;
          //Restore the EntitySet from the decrypted MemoryStream
          _mgr.CacheStateManager.RestoreCacheState(aMemoryStream, RestoreStrategy.Normal, true);
        }
        catch (Exception ex) {
          Console.WriteLine(ex.Message);
        }
        finally {
          //Close the reader and reload the Employee EntityList
          aStreamReader.Close();
          List<Employee> employees = new List<Employee>(_mgr.Employees.ToList());
        }
      }
      catch (Exception pException) {
        Console.WriteLine(pException.Message);
      }
    }

    #region Utility Methods

    private int GetCountOfEntitiesInCache() {
      return _mgr.Customers.With(QueryStrategy.CacheOnly).Count() +
        _mgr.Orders.With(QueryStrategy.CacheOnly).Count() +
        _mgr.OrderDetails.With(QueryStrategy.CacheOnly).Count();
    }

    private void PromptToContinue() {
      Console.WriteLine();
      Console.WriteLine("Press ENTER to continue...");
      Console.ReadLine();
    }

    private void DisplayCacheContents() {
      Console.WriteLine("Employees: {0}", _mgr.Employees.With(QueryStrategy.CacheOnly).Count());
      Console.WriteLine("Orders: {0}", _mgr.Orders.With(QueryStrategy.CacheOnly).Count());
      Console.WriteLine("OrderDetails: {0}", _mgr.OrderDetails.With(QueryStrategy.CacheOnly).Count());
      Console.WriteLine("Products: {0}", _mgr.Products.With(QueryStrategy.CacheOnly).Count());
      Console.WriteLine("Suppliers: {0}", _mgr.Suppliers.With(QueryStrategy.CacheOnly).Count());
      Console.WriteLine("Customers: {0}", _mgr.Customers.With(QueryStrategy.CacheOnly).Count());
    }

    public bool WriteByteArrayToFile(byte[] buff, string fileName) {
      bool response = false;

      try {
        FileStream fs = new FileStream(fileName, FileMode.Create, FileAccess.ReadWrite);
        BinaryWriter bw = new BinaryWriter(fs);
        bw.Write(buff);
        bw.Close(); //Thanks Karlo for pointing out!
        response = true;
      }
      catch (Exception ex) {
        Console.WriteLine(ex.Message);
      }

      return response;
    }

    public byte[] ReadByteArrayFromFile(string fileName) {
      byte[] buff = null;
      FileStream fs = new FileStream(fileName, FileMode.Open, FileAccess.Read);
      BinaryReader br = new BinaryReader(fs);
      long numBytes = new FileInfo(fileName).Length;
      buff = br.ReadBytes((int)numBytes);
      return buff;
    }

    #endregion Utility Methods

    #region Private Fields
    NorthwindIBEntityManager _mgr = new NorthwindIBEntityManager();
    #endregion Private Fields
  }
}
