/* A sample IIdGenerator implementation.
 * 
 *  You can use this as-is (almost) or modify as needed.  This is provided as a sample only.
 * 
 *  You may want to use an IIdGenerator to generate both temporary and real Id values for 
 *  key fields of your Entities.  Note that for tables using keys based on identity (SQL Server) 
 *  or sequence/serial values (Oracle) the StoreGeneratedIdGenerator provided by the DevForce framework should be sufficient.
 * 
 *  Possible customizations - 
 *     - change the namespace
 *     - change the class name
 *     - alter IsApplicable() logic for your data source key
 *     - alter GetNextTempId() logic for custom temporary Id values
 *     - alter GetReadIdMap() logic for custom real Id allocation
 *     
 *  Place your IdGenerator class in an assembly which can be deployed on both client and server.
 *  
 *  Note you may want to create two separate versions of your IIdGenerator, especially if you include
 *  any server-side references to assemblies not available on the client (such as EF assemblies).  
 * 
 *
*/
using System;
using System.Runtime.Serialization;

using IdeaBlade.Core;
using IdeaBlade.EntityModel;

#if !SILVERLIGHT
using System.Data;
using System.Data.Common;
#endif //!SILVERLIGHT


namespace IdeaBladeTest1Model {

  /// <summary>
  /// Sample IIdGenerator implementation.  The class should be marked with the Serializable
  /// or DataContract attribute.
  /// </summary>
  [DataContract]
  public class NumericIdGenerator : IIdGenerator {

    /// <summary>
    /// An instance of the IIdGenerator will be instantiated by the DevForce framework
    /// on both the client and server.  A public parameterless constructor is required.
    /// </summary>
    public NumericIdGenerator() {
    }

    /// <summary>
    /// Return true if this IIdGenerator should be used with the specified key.
    /// </summary>
    /// <remarks>
    /// In this example we test the key name, but this is for illustration only.  Be sure to 
    /// change this code for your implementation.  If you require an IdGenerator but DevForce
    /// can't find an "applicable" one, an exception is thrown.  Called on both client and server.
    /// </remarks>
    public bool IsApplicable(IDataSourceKey dataSourceKey) {
      //return dataSourceKey.Name.StartsWith("IdeaBladeTest", StringComparison.OrdinalIgnoreCase);
      return true;
    }

    /// <summary>
    /// Called by DevForce on the client to retrieve a temporary Id value for the specified column.
    /// </summary>
    /// <remarks>
    /// This example returns negative values starting at -1.  If you need custom temporary Ids then
    /// change this logic as necessary.
    /// Be sure to store all temporary Ids.  The framework will request them in the <see cref="TempIds"/> 
    /// property.
    /// </remarks>
    public object GetNextTempId(DataEntityProperty entityColumn) {
      Object id;
      _nextBaseId++;
      long longId = (long)(-1 * _nextBaseId);
      try {
        id = Convert.ChangeType(longId, entityColumn.DataType, System.Globalization.NumberFormatInfo.InvariantInfo);
      }
      catch {
        throw new NotSupportedException("This id generator cannot generate ids of type " + entityColumn.DataType);
      }

      _tempIds.Add(new UniqueId(entityColumn, id));
      return id;
    }

    /// <summary>
    /// Called by DevForce on the client to retrieve a collection of the temporary Id values generated since
    /// the last <see cref="Reset"/> call.  
    /// </summary>
    /// <remarks>
    /// Be sure to maintain the collection of temporary Ids generated in <see cref="GetNextTempId"/>. 
    /// </remarks>
    public UniqueIdCollection TempIds {
      get { return _tempIds; }
    }

    /// <summary>
    /// Return whether the specified Id is a temporary Id generated by this IdGenerator.
    /// </summary>
    /// <remarks>
    /// DevForce will call this on the client to test whether the specified Id is temporary and will require
    /// "fixup" as a real id.  
    /// </remarks>
    public bool IsTempId(UniqueId uniqueId) {
      if (_tempIds == null) {
        return false;
      }
      return _tempIds.Contains(uniqueId);
    }

    /// <summary>
    /// Called by DevForce after an EntityManager.SaveChanges call is performed and all temporary Id values
    /// have been "fixed up" as real Ids.  Use this to clear your temporary Id collection and reset counters.
    /// </summary>
    public void Reset() {
      _tempIds.Clear();
      _nextBaseId = 0;
    }

    /// <summary>
    /// Called by DevForce from the BOS to retrieve "real" Id values for the temporary Ids passed.
    /// </summary>
    /// <remarks>
    /// You can retrieve "real" Ids in any manner you like - this sample shows allocation
    /// of Ids from a single table holding a global sequence value.
    /// You must return a UniqueIdMap from this method containing a real Id value for each
    /// temporary Id.  Note that the UniqueId objects passed in contain information on the
    /// EntityColumn used.  You can use this information to perform more complex Id allocation
    /// if needed.
    /// Because this method is called from the server, and the server is multi-threaded,
    /// you must provide synchronization logic to handle concurrent requests.  In this
    /// sample, a lock is used within the GetNextId logic.
    /// </remarks>
    public UniqueIdMap GetRealIdMap(UniqueIdCollection tempIds, IDataSourceKey dataSourceKey) {
#if !SILVERLIGHT

      UniqueIdMap idMap = new UniqueIdMap();

      // Users might want to implement different implementations of IdGenerator on 
      // client and server if they don't want any EF references on client. 
      InitializeDbProvider(dataSourceKey);

      long nextLongId = GetNextId(tempIds.Count, dataSourceKey);
      object nextId;

      // Build a UniqueIdMap containing the mapping of temporary to real Id values.
      foreach (UniqueId aUniqueId in tempIds) {
        try {
          nextId = Convert.ChangeType(nextLongId, aUniqueId.Property.DataType);
        }
        catch {
          throw new NotSupportedException("This id generator cannot generate ids of type " + aUniqueId.Property.DataType);
        }

        idMap.Add(aUniqueId, nextId);
        nextLongId++;
      }

      return idMap;
#else
      return null;
#endif //!SILVERLIGHT

    }

#if !SILVERLIGHT

    #region Server-side custom Id allocation

    private void InitializeDbProvider(IDataSourceKey dataSourceKey) {
      // We're using simple ADO.NET functionality in this sample to directly query and update
      // the table.  You can use EF, or any other means.

      // If you want to reference IdeaBlade.EntityModel.Edm assembly you can do the following:
      //  EdmKey edmKey = (EdmKey) dataSourceKey;
      //  _connectionString = edmKey.ConnectionString;

      // If you don't want to directly reference the assembly, you can use reflection to get what
      // you need at run time (the IdeaBlade.EntityModel.Edm assembly is always loaded by the BOS).
      Type t = IdeaBlade.Core.InternalFns.GetIdeaBladeType("IdeaBlade.EntityModel.Edm.EdmKey", true);
      _connectionString = t.GetProperty("DbConnectionString").GetValue(dataSourceKey, null) as String;

      var providerName = t.GetProperty("DbProviderName").GetValue(dataSourceKey, null) as String;
      _providerFactory = DbProviderFactories.GetFactory(providerName);

      // You could also work with EF directly, or with the connectionStrings in your config file directly.
      // We don't show that here.
    }

    /// <summary>
    /// Perform allocation and return the starting next Id value.
    /// This sample disallows Id allocation for more than MaxGroupSize (1000) values
    /// at a time.  It also always allocates 100 new values at a time.
    /// These numbers are arbitrary, and for illustration only.  
    /// </summary>
    private long GetNextId(int count, IDataSourceKey dataSourceKey) {

      // Serialize access to GetNextId - the BOS is multi-threaded and concurrent saves can be in progress.
      lock (__syncLock) {

        // Do some validation before proceeding.
        TraceFns.Assert(count > 0, "GetNextId requires a positive integer as its arg.");

        TraceFns.Assert(count <= MaxGroupSize, "Cannot allocate more than "
          + MaxGroupSize + " in any one call to GetNextId");

        // Allocate more Ids only if necessary.  This allocates 100 at a time.
        if (__nextId + count > __maxNextId) {
          AllocateMoreIds(count, dataSourceKey);
        }

        long result = __nextId;
        __nextId += count;        // Next time we'll start here.
        return result;
      }
    }

    /// <summary>
    /// Allocate more "real" Id values.  This sample uses a table called NextId with
    /// "Name" and "NextId" columns.  It uses a "global" allocation strategy - that is,
    ///  Id values are not assigned on a table-specific basis, but globally within the application.
    /// </summary>
    private void AllocateMoreIds(int count, IDataSourceKey dataSourceKey) {
      const String sqlSelect = "select NextId from NextId where Name = 'GLOBAL'";
      const String sqlUpdate = "update NextId set NextId={0} where Name = 'GLOBAL' and NextId={1}";

      // Allocate the larger of the two:  amount requested or the default alloc group size.
      count = Math.Max(count, DefaultGroupSize);

      IDbConnection aConnection = CreateDbConnection();
      aConnection.Open();

      using (aConnection) {
        IDbCommand selectCommand = CreateDbCommand(aConnection);
        IDbCommand updateCommand = CreateDbCommand(aConnection);

        for (int trys = 0; trys <= MaxTrys; trys++) {

          selectCommand.CommandText = sqlSelect;
          IDataReader aDataReader = selectCommand.ExecuteReader();
          if (!aDataReader.Read()) {
            throw new ApplicationException("Unable to locate 'NextId' record");
          }

          Object tmp = aDataReader.GetValue(0);
          long nextId = (long)Convert.ChangeType(tmp, typeof(Int64));
          long newNextId = nextId + count;
          aDataReader.Close();

          // Update the table to reflect the new NextId value to be assigned next time.
          updateCommand.CommandText = String.Format(sqlUpdate, newNextId, nextId);

          // If only one record was affected - we're ok; otherwise try again.
          if (updateCommand.ExecuteNonQuery() == 1) {
            __nextId = nextId;
            __maxNextId = newNextId;
            return;
          }
        }
      }
      throw new ApplicationException("Unable to generate a new id");
    }

    private DbConnection CreateDbConnection() {
      if (String.IsNullOrEmpty(_connectionString)) {
        throw new ApplicationException("No connection string available");
      }

      DbConnection connection = _providerFactory.CreateConnection();
      connection.ConnectionString = _connectionString;
      if (connection == null) {
        throw new ApplicationException("Unable to connect using connection string: " + _connectionString);
      }
      return connection;
    }

    private DbCommand CreateDbCommand(IDbConnection connection) {
      if (connection.State != ConnectionState.Open) connection.Open();
      IDbCommand command = connection.CreateCommand();
      try {
        return (DbCommand)command;
      }
      catch {
        throw new ApplicationException("Unable to cast a DbCommand object from an IDbCommand for current connection");
      }
    }

    #endregion
#endif //!SILVERLIGHT

    // Client-side variables. Silverlight requires data members to be public.
    [DataMember]
    public UniqueIdCollection _tempIds = new UniqueIdCollection();
    [DataMember]
    public int _nextBaseId;

#if !SILVERLIGHT

    // Server-side variables.
    private static object __syncLock = new object();
    private static long __nextId = 0;
    private static long __maxNextId = 0;
    private DbProviderFactory _providerFactory;
    private String _connectionString;

    // Server-side constants
    private const int MaxTrys = 3;                // Fail after 3 unsuccessful allocation attempts.
    private const int DefaultGroupSize = 100;     // Allocate 100 new Ids at a time.
    private const int MaxGroupSize = 1000;        // Only allow 1000 Ids max in a call.

#endif //!SILVERLIGHT

  }
}
