﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using IdeaBlade.EntityModel;
using IdeaBlade.EntityModel.Server;
using IdeaBlade.Core;

namespace DomainModel {
  /// <summary>
  /// Provides server-side interception of query processing. 
  /// </summary>
  /// <remarks>
  /// You may intercept and customize query processing at several points 
  /// during the query lifecycle. The query is first authorized before 
  /// any further actions are taken: see DefaultAuthorization, 
  /// AuthorizeQuery and ClientCanQuery for information. Next the query 
  /// can be filtered to modify the query to be executed: see FilterQuery 
  /// and QueryFilters for more information. Next the query is executed: 
  /// see ExecuteQuery. Finally the query results can be authorized: see 
  /// ShouldAuthorizeQueryResult and AuthorizeQueryResult. Query processing 
  /// can be cancelled at any point in the lifecycle. 
  ///
  /// At all times during the lifecycle you have access to the Query and 
  /// the Principal for the user issuing the query. 
  ///
  /// A new instance of your EntityServerQueryInterceptor is created for 
  /// each query request. 
  /// </remarks>
  public class ServerQueryInterceptor : EntityServerQueryInterceptor {

    /// <summary>
    /// 
    /// </summary>
    /// <returns></returns>
    /// <remarks>
    /// Override this method to intercept the query to provide custom authorization. 
    /// 
    /// The default implementation calls ClientCanQuery for all types involved in 
    /// the query to determine if the user is authorized. If not authorized an 
    /// exception is thrown.
    /// </remarks>
    protected override bool AuthorizeQuery() {
      return base.AuthorizeQuery();
    }

    /// <summary>
    /// 
    /// </summary>
    /// <returns></returns>
    /// <remarks>
    /// Override to perform authorization on the results of the query. 
    /// </remarks>
    protected override bool AuthorizeQueryResult() {
      return base.AuthorizeQueryResult();
    }

    /// <summary>
    /// 
    /// </summary>
    /// <param name="type"></param>
    /// <returns></returns>
    /// <remarks>
    /// May be overridden to control which types are allowed to be queried by the user. 
    /// The base implementation tests types for the Authorization and ClientCanQuery attributes. 
    /// </remarks>
    protected override bool ClientCanQuery(Type type) {
      return base.ClientCanQuery(type);
    }

    //protected override bool DefaultAuthorization {
    //  get {
    //    return base.DefaultAuthorization;
    //  }
    //}



    /// <summary>
    /// 
    /// </summary>
    /// <remarks>
    /// Override this property to change the Default authorization result (whether 
    /// authorization succeeds if no Authorization attributes are found). 
    /// 
    /// The default is true - allowing any entity type not specifically marked 
    /// with an authorization attribute to be queried. 
    /// </remarks>
    protected override bool DefaultAuthorization {
      // Always require explicit authorization attributes on entities.
      get { return true; }
      // get { return false; }
    }

    /// <summary>
    /// 
    /// </summary>
    /// <returns></returns>
    /// <remarks>
    /// May be overridden. Insure that base.ExecuteQuery is called to insure 
    /// that the underlying query is executed. 
    /// </remarks>
    protected override bool ExecuteQuery() {
      return base.ExecuteQuery();
    }

    /// <summary>
    /// 
    /// </summary>
    /// <returns></returns>
    /// <remarks>
    /// May be overridden to perform any logic before the query is submitted. 
    /// This can include adding to the QueryFilters collection or by modifying 
    /// the query directly using the Query.Filter method or the 
    /// QueryFilterCollection classes.
    /// </remarks>
    protected override bool FilterQuery() {
      //return base.FilterQuery();

      // Add filters - let's restrict everything to UK.
      QueryFilters.AddFilter<Customer>(q => q.Where(c => c.Country == "UK"));
      QueryFilters.AddFilter<Employee>(q => q.Where(e => e.Country == "UK"));

      ////  The last statement can alternatively be expressed thusly:
      // Func<IQueryable<Employee>, IQueryable<Employee>> filter =
      //   q => q.Where(e => e.Country == "UK");
      // QueryFilters.AddFilter<Employee>(filter);

      // QueryFilters may be added to in order to restrict how a query is executed. 
      // These filters will be ignored on any queries that are not subclasses of EntityQuery. 

      return true;
    }

    protected override void OnError(Exception e, PersistenceFailure failureType) {
      TraceFns.WriteLine("Query error: " + e.Message);
      //base.OnError(e, failureType);
    }

    /// <summary>
    /// 
    /// </summary>
    /// <remarks>
    /// Whether to perform authorization by entity type on the query result. Default is false. 
    /// </remarks>
    protected override bool ShouldAuthorizeQueryResult {
      get {
        return base.ShouldAuthorizeQueryResult;
      }
    }

  }

}
