﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using IdeaBlade.EntityModel.Server;
using IdeaBlade.Core;
using IdeaBlade.EntityModel;

namespace DomainModel {

  /// <summary>
  /// Provides server-side interception of save processing. 
  /// </summary>
  /// <remarks>
  /// You may intercept and customize save processing at several points during 
  /// the save lifecycle. The save request is first authorized before any 
  /// further actions are taken: see DefaultAuthorization, AuthorizeSave and 
  /// ClientCanSave for information. Next validation can be performed on the 
  /// entities: see ValidateSave. Finally the save is executed: see ExecuteSave. 
  /// Save processing can be cancelled at any point in the lifecycle. 
  /// 
  /// At all times during the lifecycle you have access to the Principal for 
  /// the user issuing the save. 
  ///
  /// A new instance of your EntityServerSaveInterceptor is created for each save request. 

  /// </remarks>
  public class ServerSaveInterceptor : EntityServerSaveInterceptor {

    /// <summary>
    /// 
    /// </summary>
    /// <returns></returns>
    /// <remarks>
    /// May be overridden to perform any additional Authorization logic. Call 
    /// base.AuthorizeSave() internally to insure that standard authentication occurs. 
    /// 
    /// The default implementation calls ClientCanSave for all types involved in the 
    /// request to determine if the user is authorized. If not authorized an exception 
    /// is thrown.
    /// </remarks>
    protected override bool AuthorizeSave() {
      return base.AuthorizeSave();
    }

    /// <summary>
    /// 
    /// </summary>
    /// <param name="type"></param>
    /// <returns></returns>
    /// <remarks>
    /// May be overridden to control which types are allowed to be saved by the user. 
    /// The base implementation tests types for the Authorization and ClientCanSave attributes. 
    /// </remarks>
    protected override bool ClientCanSave(Type type) {
      return base.ClientCanSave(type);
    }

    //protected override bool DefaultAuthorization {
    //  get {
    //    return base.DefaultAuthorization;
    //  }
    //}


    /// <summary>
    /// 
    /// </summary>
    /// <remarks>
    /// Override this property to change the Default authorization result 
    /// (whether authorization succeeds if no Authorization attributes are found). 
    /// </remarks>
    protected override bool DefaultAuthorization {
      // Always require explicit authorization attributes on entities.
      get { return false; }
    }

    //protected override bool ExecuteSave() {
    //  return base.ExecuteSave();
    //}

    /// <summary>
    /// 
    /// </summary>
    /// <returns></returns>
    /// <remarks>
    /// May be overridden to perform any logic both immediately before and after 
    /// a save is executed. Before save operations can include adding, modifying, 
    /// deleting or removing entities from the local EntityManager. You must make 
    /// sure to call base.ExecuteSave() to insure that the underlying save does occur. 
    /// </remarks>
    protected override bool ExecuteSave() {
      // Customized save logic.

      // Pre-save logic.
      var em = EntityManager;
      var addedEmps = em.FindEntities(EntityState.Added).OfType<Employee>();

      // Mark any added emps with "Added at".
      addedEmps.ForEach(emp => emp.Notes = "Added at" + DateTime.Now);

      // Execute the save.
      var ok = base.ExecuteSave();

      // After save logic.  
      if (ok) {
        // Log any type of saved to an Employee.
        var emps = EntityManager.FindEntities(EntityState.Added | EntityState.Deleted | EntityState.Modified).OfType<Employee>();
        TraceFns.WriteLine("Emps saved: " + emps.Count());
      }

      return ok;
    }

    //protected override void OnError(Exception e, IdeaBlade.EntityModel.PersistenceFailure failureType) {
    //  base.OnError(e, failureType);
    //}

    // Log errors to any listening loggers.
    protected override void OnError(Exception e, PersistenceFailure failureType) {
      TraceFns.WriteLine("Save error: " + e.Message);
    }


    /// <summary>
    /// 
    /// </summary>
    /// <returns></returns>
    /// <remarks>
    /// May be overridden to perform any logic before the save is submitted. 
    /// The VerifierEngine used for validation is thread-safe and is shared 
    /// across all Save operations. 
    /// </remarks>
    protected override bool ValidateSave() {
      return base.ValidateSave();
    }

  }

}
