' A sample IIdGenerator implementation.
' * 
' *  You can use this as-is (almost) or modify as needed.  This is provided as a sample only.
' * 
' *  You may want to use an IIdGenerator to generate both temporary and real Id values for 
' *  key fields of your Entities.  Note that for tables using keys based on identity (SQL Server) 
' *  or sequence/serial values (Oracle) the StoreGeneratedIdGenerator provided by the DevForce framework should be sufficient.
' * 
' *  Possible customizations - 
' *     - change the namespace
' *     - change the class name
' *     - alter IsApplicable() logic for your data source key
' *     - alter GetNextTempId() logic for custom temporary Id values
' *     - alter GetReadIdMap() logic for custom real Id allocation
' *     
' *  Place your IdGenerator class in an assembly which can be deployed on both client and server.
' *  
' *  Note you may want to create two separate versions of your IIdGenerator, especially if you include
' *  any server-side references to assemblies not available on the client (such as EF assemblies).  
' * 
' *
'
Imports System.Runtime.Serialization

Imports IdeaBlade.Core
Imports IdeaBlade.EntityModel

#If Not SILVERLIGHT Then
Imports System.Data.Common
#End If '!SILVERLIGHT


Namespace IdeaBladeTest1Model

  ''' <summary>
  ''' Sample IIdGenerator implementation.  The class should be marked with the Serializable
  ''' or DataContract attribute.
  ''' </summary>
  <DataContract>
  Public Class NumericIdGenerator
	  Implements IIdGenerator

	''' <summary>
	''' An instance of the IIdGenerator will be instantiated by the DevForce framework
	''' on both the client and server.  A public parameterless constructor is required.
	''' </summary>
	Public Sub New()
	End Sub

	''' <summary>
	''' Return true if this IIdGenerator should be used with the specified key.
	''' </summary>
	''' <remarks>
	''' In this example we test the key name, but this is for illustration only.  Be sure to 
	''' change this code for your implementation.  If you require an IdGenerator but DevForce
	''' can't find an "applicable" one, an exception is thrown.  Called on both client and server.
	''' </remarks>
    Public Function IsApplicable(ByVal dataSourceKey As IDataSourceKey) As Boolean Implements IIdGenerator.IsApplicable
      'return dataSourceKey.Name.StartsWith("IdeaBladeTest", StringComparison.OrdinalIgnoreCase);
      Return True
    End Function

	''' <summary>
	''' Called by DevForce on the client to retrieve a temporary Id value for the specified column.
	''' </summary>
	''' <remarks>
	''' This example returns negative values starting at -1.  If you need custom temporary Ids then
	''' change this logic as necessary.
	''' Be sure to store all temporary Ids.  The framework will request them in the <see cref="TempIds"/> 
	''' property.
	''' </remarks>
    Public Function GetNextTempId(ByVal entityColumn As DataEntityProperty) As Object Implements IIdGenerator.GetNextTempId
      Dim id As Object
      _nextBaseId += 1
      Dim longId As Long = CLng(Fix(-1 * _nextBaseId))
      Try
        id = Convert.ChangeType(longId, entityColumn.DataType, System.Globalization.NumberFormatInfo.InvariantInfo)
      Catch
        Throw New NotSupportedException("This id generator cannot generate ids of type " & entityColumn.DataType.ToString())
      End Try

      _tempIds.Add(New UniqueId(entityColumn, id))
      Return id
    End Function

	''' <summary>
	''' Called by DevForce on the client to retrieve a collection of the temporary Id values generated since
	''' the last <see cref="Reset"/> call.  
	''' </summary>
	''' <remarks>
	''' Be sure to maintain the collection of temporary Ids generated in <see cref="GetNextTempId"/>. 
	''' </remarks>
    Public ReadOnly Property TempIds() As UniqueIdCollection Implements IIdGenerator.TempIds
      Get
        Return _tempIds
      End Get
    End Property

	''' <summary>
	''' Return whether the specified Id is a temporary Id generated by this IdGenerator.
	''' </summary>
	''' <remarks>
	''' DevForce will call this on the client to test whether the specified Id is temporary and will require
	''' "fixup" as a real id.  
	''' </remarks>
    Public Function IsTempId(ByVal uniqueId As UniqueId) As Boolean Implements IIdGenerator.IsTempId
      If _tempIds Is Nothing Then
        Return False
      End If
      Return _tempIds.Contains(uniqueId)
    End Function

	''' <summary>
	''' Called by DevForce after an EntityManager.SaveChanges call is performed and all temporary Id values
	''' have been "fixed up" as real Ids.  Use this to clear your temporary Id collection and reset counters.
	''' </summary>
    Public Sub Reset() Implements IIdGenerator.Reset
      _tempIds.Clear()
      _nextBaseId = 0
    End Sub

	''' <summary>
	''' Called by DevForce from the BOS to retrieve "real" Id values for the temporary Ids passed.
	''' </summary>
	''' <remarks>
	''' You can retrieve "real" Ids in any manner you like - this sample shows allocation
	''' of Ids from a single table holding a global sequence value.
	''' You must return a UniqueIdMap from this method containing a real Id value for each
	''' temporary Id.  Note that the UniqueId objects passed in contain information on the
	''' EntityColumn used.  You can use this information to perform more complex Id allocation
	''' if needed.
	''' Because this method is called from the server, and the server is multi-threaded,
	''' you must provide synchronization logic to handle concurrent requests.  In this
	''' sample, a lock is used within the GetNextId logic.
	''' </remarks>
    Public Function GetRealIdMap(ByVal tempIds As UniqueIdCollection, ByVal dataSourceKey As IDataSourceKey) As UniqueIdMap Implements IIdGenerator.GetRealIdMap
#If Not SILVERLIGHT Then

      Dim idMap As New UniqueIdMap()

      ' Users might want to implement different implementations of IdGenerator on 
      ' client and server if they don't want any EF references on client. 
      InitializeDbProvider(dataSourceKey)

      Dim nextLongId As Long = GetNextId(tempIds.Count, dataSourceKey)
      Dim nextId As Object

      ' Build a UniqueIdMap containing the mapping of temporary to real Id values.
      For Each aUniqueId As UniqueId In tempIds
        Try
          nextId = Convert.ChangeType(nextLongId, aUniqueId.Property.DataType)
        Catch
          Throw New NotSupportedException("This id generator cannot generate ids of type " & aUniqueId.Property.DataType.ToString())
        End Try

        idMap.Add(aUniqueId, nextId)
        nextLongId += 1
      Next aUniqueId

      Return idMap
#Else
	  Return Nothing
#End If '!SILVERLIGHT

    End Function

#If Not SILVERLIGHT Then

	#Region "Server-side custom Id allocation"

	Private Sub InitializeDbProvider(ByVal dataSourceKey As IDataSourceKey)
	  ' We're using simple ADO.NET functionality in this sample to directly query and update
	  ' the table.  You can use EF, or any other means.

	  ' If you want to reference IdeaBlade.EntityModel.Edm assembly you can do the following:
	  '  EdmKey edmKey = (EdmKey) dataSourceKey;
	  '  _connectionString = edmKey.ConnectionString;

	  ' If you don't want to directly reference the assembly, you can use reflection to get what
	  ' you need at run time (the IdeaBlade.EntityModel.Edm assembly is always loaded by the BOS).
	  Dim t As Type = IdeaBlade.Core.InternalFns.GetIdeaBladeType("IdeaBlade.EntityModel.Edm.EdmKey", True)
	  _connectionString = TryCast(t.GetProperty("DbConnectionString").GetValue(dataSourceKey, Nothing), String)

	  Dim providerName = TryCast(t.GetProperty("DbProviderName").GetValue(dataSourceKey, Nothing), String)
	  _providerFactory = DbProviderFactories.GetFactory(providerName)

	  ' You could also work with EF directly, or with the connectionStrings in your config file directly.
	  ' We don't show that here.
	End Sub

	''' <summary>
	''' Perform allocation and return the starting next Id value.
	''' This sample disallows Id allocation for more than MaxGroupSize (1000) values
	''' at a time.  It also always allocates 100 new values at a time.
	''' These numbers are arbitrary, and for illustration only.  
	''' </summary>
	Private Function GetNextId(ByVal count As Integer, ByVal dataSourceKey As IDataSourceKey) As Long

	  ' Serialize access to GetNextId - the BOS is multi-threaded and concurrent saves can be in progress.
	  SyncLock __syncLock

		' Do some validation before proceeding.
		TraceFns.Assert(count > 0, "GetNextId requires a positive integer as its arg.")

		TraceFns.Assert(count <= MaxGroupSize, "Cannot allocate more than " & MaxGroupSize & " in any one call to GetNextId")

		' Allocate more Ids only if necessary.  This allocates 100 at a time.
		If __nextId + count > __maxNextId Then
		  AllocateMoreIds(count, dataSourceKey)
		End If

		Dim result As Long = __nextId
		__nextId += count ' Next time we'll start here.
		Return result
	  End SyncLock
	End Function

	''' <summary>
	''' Allocate more "real" Id values.  This sample uses a table called NextId with
	''' "Name" and "NextId" columns.  It uses a "global" allocation strategy - that is,
	'''  Id values are not assigned on a table-specific basis, but globally within the application.
	''' </summary>
	Private Sub AllocateMoreIds(ByVal count As Integer, ByVal dataSourceKey As IDataSourceKey)
	  Const sqlSelect As String = "select NextId from NextId where Name = 'GLOBAL'"
	  Const sqlUpdate As String = "update NextId set NextId={0} where Name = 'GLOBAL' and NextId={1}"

	  ' Allocate the larger of the two:  amount requested or the default alloc group size.
	  count = Math.Max(count, DefaultGroupSize)

	  Dim aConnection As IDbConnection = CreateDbConnection()
	  aConnection.Open()

	  Using aConnection
		Dim selectCommand As IDbCommand = CreateDbCommand(aConnection)
		Dim updateCommand As IDbCommand = CreateDbCommand(aConnection)

		For trys As Integer = 0 To MaxTrys

		  selectCommand.CommandText = sqlSelect
		  Dim aDataReader As IDataReader = selectCommand.ExecuteReader()
		  If Not aDataReader.Read() Then
			Throw New ApplicationException("Unable to locate 'NextId' record")
		  End If

		  Dim tmp As Object = aDataReader.GetValue(0)
		  Dim nextId As Long = CLng(Fix(Convert.ChangeType(tmp, GetType(Int64))))
		  Dim newNextId As Long = nextId + count
		  aDataReader.Close()

		  ' Update the table to reflect the new NextId value to be assigned next time.
		  updateCommand.CommandText = String.Format(sqlUpdate, newNextId, nextId)

		  ' If only one record was affected - we're ok; otherwise try again.
		  If updateCommand.ExecuteNonQuery() = 1 Then
			__nextId = nextId
			__maxNextId = newNextId
			Return
		  End If
		Next trys
	  End Using
	  Throw New ApplicationException("Unable to generate a new id")
	End Sub

	Private Function CreateDbConnection() As DbConnection
	  If String.IsNullOrEmpty(_connectionString) Then
		Throw New ApplicationException("No connection string available")
	  End If

	  Dim connection As DbConnection = _providerFactory.CreateConnection()
	  connection.ConnectionString = _connectionString
	  If connection Is Nothing Then
		Throw New ApplicationException("Unable to connect using connection string: " & _connectionString)
	  End If
	  Return connection
	End Function

	Private Function CreateDbCommand(ByVal connection As IDbConnection) As DbCommand
	  If connection.State <> ConnectionState.Open Then
		  connection.Open()
	  End If
	  Dim command As IDbCommand = connection.CreateCommand()
	  Try
		Return CType(command, DbCommand)
	  Catch
		Throw New ApplicationException("Unable to cast a DbCommand object from an IDbCommand for current connection")
	  End Try
	End Function

	#End Region
#End If '!SILVERLIGHT

	' Client-side variables. Silverlight requires data members to be public.
	<DataMember>
	Public _tempIds As New UniqueIdCollection()
	<DataMember>
	Public _nextBaseId As Integer

#If Not SILVERLIGHT Then

	' Server-side variables.
	Private Shared __syncLock As New Object()
	Private Shared __nextId As Long = 0
	Private Shared __maxNextId As Long = 0
	Private _providerFactory As DbProviderFactory
	Private _connectionString As String

	' Server-side constants
	Private Const MaxTrys As Integer = 3 ' Fail after 3 unsuccessful allocation attempts.
	Private Const DefaultGroupSize As Integer = 100 ' Allocate 100 new Ids at a time.
	Private Const MaxGroupSize As Integer = 1000 ' Only allow 1000 Ids max in a call.

#End If '!SILVERLIGHT

  End Class
End Namespace
