﻿Imports System.Text

Imports IdeaBlade.EntityModel
Imports IdeaBlade.EntityModel.Server
Imports IdeaBlade.Core

Namespace DomainModel
  ''' <summary>
  ''' Provides server-side interception of query processing. 
  ''' </summary>
  ''' <remarks>
  ''' You may intercept and customize query processing at several points 
  ''' during the query lifecycle. The query is first authorized before 
  ''' any further actions are taken: see DefaultAuthorization, 
  ''' AuthorizeQuery and ClientCanQuery for information. Next the query 
  ''' can be filtered to modify the query to be executed: see FilterQuery 
  ''' and QueryFilters for more information. Next the query is executed: 
  ''' see ExecuteQuery. Finally the query results can be authorized: see 
  ''' ShouldAuthorizeQueryResult and AuthorizeQueryResult. Query processing 
  ''' can be cancelled at any point in the lifecycle. 
  '''
  ''' At all times during the lifecycle you have access to the Query and 
  ''' the Principal for the user issuing the query. 
  '''
  ''' A new instance of your EntityServerQueryInterceptor is created for 
  ''' each query request. 
  ''' </remarks>
  Public Class ServerQueryInterceptor
	  Inherits EntityServerQueryInterceptor

	''' <summary>
	''' 
	''' </summary>
	''' <returns></returns>
	''' <remarks>
	''' Override this method to intercept the query to provide custom authorization. 
	''' 
	''' The default implementation calls ClientCanQuery for all types involved in 
	''' the query to determine if the user is authorized. If not authorized an 
	''' exception is thrown.
	''' </remarks>
	Protected Overrides Function AuthorizeQuery() As Boolean
	  Return MyBase.AuthorizeQuery()
	End Function

	''' <summary>
	''' 
	''' </summary>
	''' <returns></returns>
	''' <remarks>
	''' Override to perform authorization on the results of the query. 
	''' </remarks>
	Protected Overrides Function AuthorizeQueryResult() As Boolean
	  Return MyBase.AuthorizeQueryResult()
	End Function

	''' <summary>
	''' 
	''' </summary>
	''' <param name="type"></param>
	''' <returns></returns>
	''' <remarks>
	''' May be overridden to control which types are allowed to be queried by the user. 
	''' The base implementation tests types for the Authorization and ClientCanQuery attributes. 
	''' </remarks>
	Protected Overrides Function ClientCanQuery(ByVal type As Type) As Boolean
	  Return MyBase.ClientCanQuery(type)
	End Function

	'protected override bool DefaultAuthorization {
	'  get {
	'    return base.DefaultAuthorization;
	'  }
	'}



	''' <summary>
	''' 
	''' </summary>
	''' <remarks>
	''' Override this property to change the Default authorization result (whether 
	''' authorization succeeds if no Authorization attributes are found). 
	''' 
	''' The default is true - allowing any entity type not specifically marked 
	''' with an authorization attribute to be queried. 
	''' </remarks>
	Protected Overrides ReadOnly Property DefaultAuthorization() As Boolean
	  ' Always require explicit authorization attributes on entities.
	  Get
		  Return True
	  End Get
	  ' get { return false; }
	End Property

	''' <summary>
	''' 
	''' </summary>
	''' <returns></returns>
	''' <remarks>
	''' May be overridden. Insure that base.ExecuteQuery is called to insure 
	''' that the underlying query is executed. 
	''' </remarks>
	Protected Overrides Function ExecuteQuery() As Boolean
	  Return MyBase.ExecuteQuery()
	End Function

	''' <summary>
	''' 
	''' </summary>
	''' <returns></returns>
	''' <remarks>
	''' May be overridden to perform any logic before the query is submitted. 
	''' This can include adding to the QueryFilters collection or by modifying 
	''' the query directly using the Query.Filter method or the 
	''' QueryFilterCollection classes.
	''' </remarks>
	Protected Overrides Function FilterQuery() As Boolean
	  'return base.FilterQuery();

	  ' Add filters - let's restrict everything to UK.
	  QueryFilters.AddFilter(Of Customer)(Function(q) q.Where(Function(c) c.Country = "UK"))
	  QueryFilters.AddFilter(Of Employee)(Function(q) q.Where(Function(e) e.Country = "UK"))

	  '//  The last statement can alternatively be expressed thusly:
	  ' Func<IQueryable<Employee>, IQueryable<Employee>> filter =
	  '   q => q.Where(e => e.Country == "UK");
	  ' QueryFilters.AddFilter<Employee>(filter);

	  ' QueryFilters may be added to in order to restrict how a query is executed. 
	  ' These filters will be ignored on any queries that are not subclasses of EntityQuery. 

	  Return True
	End Function

	Protected Overrides Sub OnError(ByVal e As Exception, ByVal failureType As PersistenceFailure)
	  TraceFns.WriteLine("Query error: " & e.Message)
	  'base.OnError(e, failureType);
	End Sub

	''' <summary>
	''' 
	''' </summary>
	''' <remarks>
	''' Whether to perform authorization by entity type on the query result. Default is false. 
	''' </remarks>
	Protected Overrides ReadOnly Property ShouldAuthorizeQueryResult() As Boolean
	  Get
		Return MyBase.ShouldAuthorizeQueryResult
	  End Get
	End Property

  End Class

End Namespace
