﻿Imports System.Text

Imports IdeaBlade.EntityModel.Server
Imports IdeaBlade.Core
Imports IdeaBlade.EntityModel

Namespace DomainModel

  ''' <summary>
  ''' Provides server-side interception of save processing. 
  ''' </summary>
  ''' <remarks>
  ''' You may intercept and customize save processing at several points during 
  ''' the save lifecycle. The save request is first authorized before any 
  ''' further actions are taken: see DefaultAuthorization, AuthorizeSave and 
  ''' ClientCanSave for information. Next validation can be performed on the 
  ''' entities: see ValidateSave. Finally the save is executed: see ExecuteSave. 
  ''' Save processing can be cancelled at any point in the lifecycle. 
  ''' 
  ''' At all times during the lifecycle you have access to the Principal for 
  ''' the user issuing the save. 
  '''
  ''' A new instance of your EntityServerSaveInterceptor is created for each save request. 

  ''' </remarks>
  Public Class ServerSaveInterceptor
	  Inherits EntityServerSaveInterceptor

	''' <summary>
	''' 
	''' </summary>
	''' <returns></returns>
	''' <remarks>
	''' May be overridden to perform any additional Authorization logic. Call 
	''' base.AuthorizeSave() internally to insure that standard authentication occurs. 
	''' 
	''' The default implementation calls ClientCanSave for all types involved in the 
	''' request to determine if the user is authorized. If not authorized an exception 
	''' is thrown.
	''' </remarks>
	Protected Overrides Function AuthorizeSave() As Boolean
	  Return MyBase.AuthorizeSave()
	End Function

	''' <summary>
	''' 
	''' </summary>
	''' <param name="type"></param>
	''' <returns></returns>
	''' <remarks>
	''' May be overridden to control which types are allowed to be saved by the user. 
	''' The base implementation tests types for the Authorization and ClientCanSave attributes. 
	''' </remarks>
	Protected Overrides Function ClientCanSave(ByVal type As Type) As Boolean
	  Return MyBase.ClientCanSave(type)
	End Function

	'protected override bool DefaultAuthorization {
	'  get {
	'    return base.DefaultAuthorization;
	'  }
	'}


	''' <summary>
	''' 
	''' </summary>
	''' <remarks>
	''' Override this property to change the Default authorization result 
	''' (whether authorization succeeds if no Authorization attributes are found). 
	''' </remarks>
	Protected Overrides ReadOnly Property DefaultAuthorization() As Boolean
	  ' Always require explicit authorization attributes on entities.
	  Get
		  Return False
	  End Get
	End Property

	'protected override bool ExecuteSave() {
	'  return base.ExecuteSave();
	'}

	''' <summary>
	''' 
	''' </summary>
	''' <returns></returns>
	''' <remarks>
	''' May be overridden to perform any logic both immediately before and after 
	''' a save is executed. Before save operations can include adding, modifying, 
	''' deleting or removing entities from the local EntityManager. You must make 
	''' sure to call base.ExecuteSave() to insure that the underlying save does occur. 
	''' </remarks>
	Protected Overrides Function ExecuteSave() As Boolean
	  ' Customized save logic.

	  ' Pre-save logic.
	  Dim em = EntityManager
	  Dim addedEmps = em.FindEntities(EntityState.Added).OfType(Of Employee)()

	  ' Mark any added emps with "Added at".
	  addedEmps.ForEach(Function(emp) emp.Notes = "Added at" & Date.Now)

	  ' Execute the save.
	  Dim ok = MyBase.ExecuteSave()

	  ' After save logic.  
	  If ok Then
		' Log any type of saved to an Employee.
		Dim emps = EntityManager.FindEntities(EntityState.Added Or EntityState.Deleted Or EntityState.Modified).OfType(Of Employee)()
		TraceFns.WriteLine("Emps saved: " & emps.Count())
	  End If

	  Return ok
	End Function

	'protected override void OnError(Exception e, IdeaBlade.EntityModel.PersistenceFailure failureType) {
	'  base.OnError(e, failureType);
	'}

	' Log errors to any listening loggers.
	Protected Overrides Sub OnError(ByVal e As Exception, ByVal failureType As PersistenceFailure)
	  TraceFns.WriteLine("Save error: " & e.Message)
	End Sub


	''' <summary>
	''' 
	''' </summary>
	''' <returns></returns>
	''' <remarks>
	''' May be overridden to perform any logic before the save is submitted. 
	''' The VerifierEngine used for validation is thread-safe and is shared 
	''' across all Save operations. 
	''' </remarks>
	Protected Overrides Function ValidateSave() As Boolean
	  Return MyBase.ValidateSave()
	End Function

  End Class

End Namespace
